library KillCopy;

(*
Description: Packer plug-in interface to KillCopy for Total Commander
Author: Kovcs Endre Jnos
E-Mail: kovacsendre@coder.hu

TODO: filter file list. KillCopy copies folders recursively

Changes: ProgMan13
*)

uses
  Messages,
  Windows,
  wcxhead;

{$E wcx}
{$IFDEF WIN64}
  {$E w64}
{$ENDIF}

{$R *.res}

const
  szKCKey = 'SOFTWARE\KILLSOFT\KillCopy';

var
  AppPath: String;
  PackerChangeVolProc:TChangeVolProcW;
  PackerProcessDataProc:TProcessDataProcW;

{
function StrEnd(const Str: PChar): PChar; assembler;
(* clipped from Borland's SysUtils.pas to reduce code size *)
asm
        MOV     EDX,EDI
        MOV     EDI,EAX
        MOV     ECX,0FFFFFFFFH
        XOR     AL,AL
        REPNE   SCASB
        LEA     EAX,[EDI-1]
        MOV     EDI,EDX
end;
}
function StrEnd(const Str: PWideChar): PWideChar;
begin
  Result := Str;
  while Result^ <> #0 do
    Inc(Result);
end;

function ExtractFilePath(const FileName: string): string;
var
  i: Integer;
begin
  i := Length(FileName);
  while i > 0 do begin
    if (FileName[i] = '\') or (FileName[i] = ':') then Break;
    dec(i)
  end;

  if i > 0 then
    Result := Copy(FileName, 1, i)
  else
    Result := '';
end;

function ExpandEnvStrings(const Str: string): string;
var
  Len: integer;
begin
  Result := '';
  Len := ExpandEnvironmentStrings(PChar(Str), nil, 0);
  if Len > 0 then
  begin
    SetLength(Result, Len - 1);
    ExpandEnvironmentStrings(PChar(Str), PChar(Result), Len);
  end;
end;

function GetRegString(const KeyName, ValueName: string; RootKey: HKEY = HKEY_CLASSES_ROOT): String;
var
  Key: HKey;
  StrLen: Integer;
  Buffer: array[0..1023] of Char;
begin
  Result := '';
  if RegOpenKeyEx(RootKey,PChar(KeyName),0,KEY_READ,Key) = ERROR_SUCCESS then begin
    StrLen := SizeOf(Buffer);
    if RegQueryValueEx(Key, PChar(ValueName), nil, nil, @Buffer, @StrLen) = ERROR_SUCCESS then
      Result := Buffer;
    RegCloseKey(Key);
  end
end;

function GetSystemDir: String;
var
  Buf: array[0..MAX_PATH] of Char;
begin
  GetSystemDirectory(Buf, SizeOf(Buf));
  Result := Buf;
end;

function GetTempName: String;
var
  TempPath, TempName: array[0..MAX_PATH] of Char;
begin
  FillChar(TempPath, SizeOf(TempPath), 0);
  FillChar(TempName, SizeOf(TempName), 0);
  GetTempPath(MAX_PATH, TempPath);
  GetTempFileName(TempPath,'kcp',0,TempName);
  Result := TempName;
end;

function ReadINIString(const FileName, Section, Ident, Default: string): string;
var
  Buffer: array[0..2047] of Char;
begin
  SetString(Result, Buffer, GetPrivateProfileString(PChar(Section),
    PChar(Ident), PChar(Default), Buffer, Length(Buffer), PChar(FileName)));
end;

function GetModuleName(Module: HMODULE): string;
var
  ModName: array[0..MAX_PATH] of Char;
begin
  SetString(Result, ModName, GetModuleFileName(Module, ModName, Length(ModName)));
end;

(*************************************************)

function OpenArchive(var ArchiveData:TOpenArchiveData):THandle; stdcall;
begin
  Result := 0;
  ArchiveData.OpenResult := E_NOT_SUPPORTED;
end;

function OpenArchiveW(var ArchiveData:TOpenArchiveDataW):THandle; stdcall;
begin
  Result := 0;
  ArchiveData.OpenResult := E_NOT_SUPPORTED;
end;

function CloseArchive(hArcData:thandle):Integer; stdcall;
begin
  Result:=0;
end;

function ReadHeader(hArcData:thandle;var HeaderData:THeaderData):Integer; stdcall;
begin
  Result := E_NOT_SUPPORTED;
end;

function ReadHeaderExW(hArcData: THandle; var HD: tHeaderDataExW): Integer; stdcall;
begin
  Result := E_NOT_SUPPORTED;
end;

function ProcessFile(hArcData:thandle;Operation:longint;DestPath,DestName:pansichar):Integer; stdcall;
begin
  Result := E_NOT_SUPPORTED;
end;

function ProcessFileW(hArcData:thandle;Operation:longint;DestPath,DestName:pchar):Integer; stdcall;
begin
  Result := E_NOT_SUPPORTED;
end;

function GetPackerCaps:integer; stdcall;
begin
  Result := PK_CAPS_NEW or PK_CAPS_MODIFY or PK_CAPS_MULTIPLE or PK_CAPS_HIDE;
end;

function ExecKillCopy(const ListFileName, DestFolder: String; Flags: integer):Boolean;
var
  CommandLine: String;
  StartupInfo: TStartupInfo;
  ProcessInfo: TProcessInformation;
  IniFile, FullPath: string;
begin
  Result := False;
  FullPath := '';

  FillChar(StartupInfo, SizeOf(StartupInfo), 0);
  StartupInfo.cb := SizeOf(StartupInfo);

  INIFile := ExtractFilePath(GetModuleName(HInstance)) + 'KillCopy.ini';
  FullPath := ReadINIString(IniFile, 'KillCopy', 'FullPath', FullPath);
  if FullPath = '' then
    FullPath := AppPath + 'KillCopy.exe';

  CommandLine := '"' + FullPath + '" "|' + ListFileName + '|' + DestFolder + '|';
  //CommandLine := '"' + AppPath + 'KillCopy.exe" "|' + ListFileName + '|' + DestFolder + '|';

  if (Flags and PK_PACK_MOVE_FILES) <> 0 then
    CommandLine := CommandLine + ' -m"'
  else
    CommandLine := CommandLine + '"';

  if CreateProcess(nil, PChar(CommandLine), nil, nil, False, 0, nil, PChar(AppPath), StartupInfo, ProcessInfo) then begin
    CloseHandle(ProcessInfo.hThread); // Close thread handle
//    WaitForSingleObject(ProcessInfo.hProcess, INFINITE); // Wait until child process exits
    CloseHandle(ProcessInfo.hProcess); // Close process handle
    Result := True; //All OK
  end
end;

function PackFiles(PackedFile,SubPath,SrcPath,AddList:PAnsiChar;Flags:integer):integer; stdcall;
begin
  Result := 0;
end;

function PackFilesW(PackedFile,SubPath,SrcPath,AddList:PChar;Flags:integer):integer; stdcall;
var
  FileName, ListFileName, List, LastDir: String;
  ListFile: NativeUInt;
  p: PChar;
  BytesWritten: Cardinal;
begin
  Result := 0;
  LastDir := '';
  p := AddList;

  // --- Build list file ---

  List := '';

  //source files
  while (p^ <> #0) do begin
    FileName := String(SrcPath) + String(p);
    if (FileName <> LastDir) and (Length(FileName) > 0) and (FileName[Length(FileName)] = '\') then
      if (Pos(LastDir, FileName) <> 1) then
        LastDir := FileName;
    if (LastDir = FileName) or (LastDir = '') then
      List := List + FileName + #13#10
    else if (LastDir <> '') and (Pos(LastDir, FileName) <> 1) then
      List := List + FileName + #13#10;
    p := StrEnd(p)+1;
  end;

  if Length(List) > 0 then begin
    // --- Create list file ---
    ListFileName := GetTempName;
    ListFile := CreateFile(PChar(ListFileName),
        GENERIC_WRITE, FILE_SHARE_READ, nil,
        CREATE_ALWAYS, FILE_ATTRIBUTE_ARCHIVE, 0);

    WriteFile(ListFile, AnsiString(List)[1], Length(AnsiString(List)), BytesWritten, nil);

    CloseHandle(ListFile);

    //MessageBox(GetActiveWindow, PChar(List), 'KillCopy plugin', MB_OK or MB_SETFOREGROUND);

    // --- Execute KillCopy ---
    if not ExecKillCopy(ListFileName, ExtractFilePath(PackedFile), Flags) then
      Result := E_EOPEN;
  end
  else //List is empty, nothing to do?! Unlikely... report an error
    Result := E_EABORTED;
end;

procedure SetChangeVolProc(hArcData:THandle;ChangeVolProc:TChangeVolProc); stdcall;
begin
  //PackerChangeVolProc:=ChangeVolProc;
end;

procedure SetProcessDataProc(hArcData:THandle;ProcessDataProc:TProcessDataProc); stdcall;
begin
  //PackerProcessDataProc:=ProcessDataProc;
end;

procedure SetProcessDataProcW(hArcData:THandle;ProcessDataProc:TProcessDataProcW); stdcall;
begin
  PackerProcessDataProc:=ProcessDataProc;
end;

function DeleteFiles(PackedFile,DeleteList:pansichar):integer; stdcall;
begin
  result:=E_NOT_SUPPORTED;
end;

function DeleteFilesW(PackedFile,DeleteList:pchar):integer; stdcall;
begin
  result:=E_NOT_SUPPORTED;
end;

exports
  OpenArchiveW,
  OpenArchive,
  CloseArchive,
  ReadHeader,
  ReadHeaderExW,
  ProcessFile,
  ProcessFileW,
  GetPackerCaps,
  PackFilesW,
  PackFiles,
  DeleteFiles,
  DeleteFilesW,
  SetChangeVolProc,
  SetProcessDataProc,
  SetProcessDataProcW;

begin
  //Get KillCopy's application path (Windows System folder)
  AppPath := GetSystemDir;
  if AppPath = '' then //App path is still unknown
  {TODO: Prompt user to enter the path to KillCopy}
  else
    AppPath := AppPath + '\';

//  InstPath := GetRegString(szKCKey, 'InstallPath', HKEY_LOCAL_MACHINE);
end.
